"""Objects that can be handled by documentation builder class."""

from dataclasses import dataclass
from typing import List
from pathlib import Path


@dataclass
class ParsedDocstring:
    """
    Parsed docstring from module, class or function.

    If uses style as rst or markdown will be also
    prepaired.
    """

    raw: str

    def __post_init__(self):
        """Render markdown or ReST  docstring."""
        # TODO process markdown or rst in docstring
        self.doc = self.raw


@dataclass
class ParsedImport:
    """Parsed import from module."""

    import_from: str
    import_import: List[str]
    import_as: List[str]


@dataclass
class ParsedStatement:
    """Parsed statement from class or module."""

    name: str
    value: str
    annotation: str


@dataclass
class ParsedParameter:
    """Parsed parameter from class, method or funcion definition."""

    name: str
    annotation: str
    default: str


@dataclass
class ParsedArgument:
    """Parsed argument from call callable."""

    value: str
    name: str


@dataclass
class ParsedDecorator:
    """Parsed decorator from class, method of function."""

    name: str
    factory_parameters: List[ParsedArgument]


@dataclass
class ParsedFunction:
    """Parsed function or method."""

    name: str
    docstring: ParsedDocstring
    paramenters: List[ParsedParameter]
    decorators: List[ParsedDecorator]
    returns: str


@dataclass
class ParsedClass:
    """Parsed class."""

    name: str
    docstring: ParsedDocstring
    parent_class: List[ParsedArgument]
    decorators: List[ParsedDecorator]
    variables: List[ParsedStatement]
    methods: List[ParsedFunction]


@dataclass
class ParsedModule:
    """Parsed module."""

    name: str
    path: Path
    docstring: ParsedDocstring
    imports: List[ParsedImport]
    statements: List[ParsedStatement]
    classes: List[ParsedClass]
    functions: List[ParsedFunction]



"""Base documentation builder class."""

from pathlib import Path
from abc import ABC, abstractmethod
from .parsed_objects import ParsedModule


class BaseBuilder(ABC):
    """Abstract class for converting ParsedModule to a special style string."""

    path: Path
    module: ParsedModule
    text: str

    def __init__(self, module: ParsedModule):
        """Create builder for parsed module."""
        self.module = module

    @abstractmethod
    def feed(self):
        """Convert ParsedModule to a string and stores it in self.text."""
        self.text = ''

    def save(self):
        """Save file at location specified on init."""
        with open(self.path.absolute(), 'w') as file:
            file.write(self.text)



"""Parser based on parso. Returns objects compatible with BaseBuilder class."""

from pathlib import Path
from typing import List, Iterator, Union
import parso
from parso.python.tree import Node, Module, Class, Function, Keyword,\
    PythonNode, Name, Operator
from .parsed_objects import ParsedArgument, ParsedClass, ParsedDecorator,\
    ParsedDocstring, ParsedFunction, ParsedImport, ParsedModule,\
    ParsedParameter, ParsedStatement


def extract_doc(node: Union[Module, Class, Function]) -> ParsedDocstring:
    doc_node = node.get_doc_node()
    if doc_node:
        doc = eval(doc_node.value)
        # eval function is way to extract docstring exacly as anyone want to
        # injections cannot be here, because parso search only for docstring
    else:
        doc = ''
    return ParsedDocstring(doc.strip())


def extract_imports(node: Module) -> List[ParsedImport]:
    node_imports = []
    for node_import in node.iter_imports():
        import_data = dict(import_from=[],
                           import_import=[],
                           import_as=[])
        scan = 'from'
        for part_node in node_import.children:
            if isinstance(part_node, Operator):
                continue
            if isinstance(part_node, Keyword):
                keyword = part_node.value
                if keyword in ['from', 'import', 'as']:
                    scan = keyword
            else:
                if isinstance(part_node, Name):
                    import_data['import_' + scan].append(part_node.value)
                elif isinstance(part_node, PythonNode):
                    value = ['']
                    for child in part_node.children:
                        if isinstance(child, Operator):
                            if child.value.strip() == '.':
                                value[-1] += '.'
                            elif child.value.strip() == ',':
                                value.append('')
                            else:
                                import_data['import_' + scan].append(
                                    part_node.get_code())
                                continue
                        else:
                            value[-1] += child.value
                    import_data['import_' + scan].extend(value)
                else:
                    import_data['import_' + scan].append(part_node.get_code())
        node_imports.append(ParsedImport(**import_data))
    return node_imports


def extract_statements(node: Node) -> List[ParsedStatement]:
    return []


def extract_params(node: Node) -> List[ParsedParameter]:
    return []


def extract_functions(node: Node) -> List[ParsedFunction]:
    return []


def extract_classes(node: Node) -> List[ParsedClass]:
    return []


def extract(file_name: Path) -> ParsedModule:
    with open(file_name.absolute()) as file:
        root_node = parso.parse(file.read())
    return ParsedModule(name=file_name.name[:-3],
                        path=file_name,
                        docstring=extract_doc(root_node),
                        imports=extract_imports(root_node),
                        statements=extract_statements(root_node),
                        classes=extract_classes(root_node),
                        functions=extract_functions(root_node))


def find_python(base: Path) -> Iterator[Path]:
    if base.is_dir():
        for child in base.iterdir():
            for python_file in find_python(child):
                yield python_file
    elif base.match('*.py'):
        yield base


def find_and_extract(base: Path) -> Iterator[ParsedModule]:
    for python_file in find_python(base):
        yield extract(python_file)


def test_find():
    """Test for recursive python files search."""
    module_dir = Path('testset/mypy')
    paths = list(find_python(module_dir))
    assert paths


def test_find_and_extract():
    """Test for recursive files parsing."""
    import itertools
    module_dir = Path('testset/mypy')
    parsed_modules = list(itertools.islice(find_and_extract(module_dir), 19))
    assert parsed_modules
    breakpoint()



from setuptools import setup, find_packages
from os import path
import json


def requirements_from_pipfile_lock(file_path='Pipfile.lock'):
    with open(file_path) as file:
        pipfile = json.load(file)
    return [r + a['version'] for r, a in pipfile['default'].items()]


requirements = requirements_from_pipfile_lock()
here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as file:
    long_description = file.read()


setup(
    name='moduledocs',
    version='0.0.1',
    description='Static documentation extraction tool for python language',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/machineandme/moduledocs',
    author='Machine & Me',
    author_email='ceo@machineand.me',
    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 3 - Alpha',

        # Indicate who your project is intended for
        'Intended Audience :: Developers',
        'Topic :: Documentation',

        # Pick your license as you wish
        'License :: OSI Approved :: MIT License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        # These classifiers are *not* checked by 'pip install'. See instead
        # 'python_requires' below.
        'Programming Language :: Python :: 3',
    ],
    keywords='generate documentation markdown html static parser',
    packages=find_packages(exclude=['contrib', 'docs', 'tests', 'examples']),
    install_requires=requirements,
    python_requires='>=3.5.*',
    entry_points={
        'console_scripts': [
            'moduledocs=moduledocs.cli:main',
        ],
    },
    project_urls={
        'Bug Reports': 'https://github.com/machineandme/moduledocs/issues',
        'Funding': 'https://www.patreon.com/machineandme',
        'Source': 'https://github.com/machineandme/moduledocs'
    },
)
